/*
 * Copyright (c) 2021 Arm Limited.
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include "pooling.hpp"
#include <cstdint>

#if defined(ARM_COMPUTE_ENABLE_SVE) && defined(ARM_COMPUTE_ENABLE_SVE2)

namespace arm_conv {
namespace pooling {


void sve_u8q_nhwc_max_generic_depthfirst_impl(
  const uint64_t,
  const uint64_t n_valid_cells,
  uint64_t n_channels,
  const uint8_t *const *const inptrs,
  uint8_t *outptr,
  const Requantize32 &qp
)
{
  __asm__ __volatile__(
    "ptrue p4.b\n"
    "mov x28, #0x0\n"
    "cntb x27\n"
    "cntb x26, ALL, MUL #2\n"
    "cntb x25, ALL, MUL #3\n"
    "whilelt p3.b, x28, %x[n_channels]\n"
    "whilelt p2.b, x27, %x[n_channels]\n"
    "whilelt p1.b, x26, %x[n_channels]\n"
    "whilelt p0.b, x25, %x[n_channels]\n"
    "b.none 7f\n"
    "1:"  // 4-vectors of channels
    "mov z10.b, #0x0\n"
    "mov x19, %x[inptrs]\n"
    "mov z9.b, #0x0\n"
    "lsr x24, %x[n_valid_cells], #0x2\n"
    "mov z8.b, #0x0\n"
    "mov z7.b, #0x0\n"
    "cbz x24, 4f\n"
    "ldp x23, x22, [x19, #0x0]\n"
    "ldp x21, x20, [x19, #0x10]\n"
    "add x19, x19, #0x20\n"
    "subs x24, x24, #0x1\n"
    "ld1b { z3.b }, p3/Z, [x23, x28]\n"
    "ld1b { z2.b }, p3/Z, [x22, x28]\n"
    "ld1b { z1.b }, p3/Z, [x21, x28]\n"
    "ld1b { z0.b }, p3/Z, [x20, x28]\n"
    "ld1b { z31.b }, p2/Z, [x23, x27]\n"
    "ld1b { z30.b }, p2/Z, [x22, x27]\n"
    "ld1b { z22.b }, p2/Z, [x21, x27]\n"
    "ld1b { z29.b }, p2/Z, [x20, x27]\n"
    "ld1b { z28.b }, p1/Z, [x23, x26]\n"
    "ld1b { z27.b }, p1/Z, [x22, x26]\n"
    "ld1b { z21.b }, p1/Z, [x21, x26]\n"
    "ld1b { z26.b }, p1/Z, [x20, x26]\n"
    "ld1b { z16.b }, p0/Z, [x23, x25]\n"
    "ld1b { z25.b }, p0/Z, [x22, x25]\n"
    "ld1b { z20.b }, p0/Z, [x21, x25]\n"
    "ld1b { z24.b }, p0/Z, [x20, x25]\n"
    "beq 3f\n"
    "2:"  // 4-vectors of channels: 4 inputs loop
    "movprfx z19, z3\n umax z19.b, p4/M, z19.b, z2.b\n"
    "ldp x23, x22, [x19, #0x0]\n"
    "subs x24, x24, #0x1\n"
    "movprfx z23, z1\n umax z23.b, p4/M, z23.b, z0.b\n"
    "ldp x21, x20, [x19, #0x10]\n"
    "add x19, x19, #0x20\n"
    "movprfx z18, z31\n umax z18.b, p4/M, z18.b, z30.b\n"
    "ld1b { z3.b }, p3/Z, [x23, x28]\n"
    "umax z22.b, p4/M, z22.b, z29.b\n"
    "movprfx z17, z28\n umax z17.b, p4/M, z17.b, z27.b\n"
    "ld1b { z2.b }, p3/Z, [x22, x28]\n"
    "umax z21.b, p4/M, z21.b, z26.b\n"
    "ld1b { z1.b }, p3/Z, [x21, x28]\n"
    "umax z16.b, p4/M, z16.b, z25.b\n"
    "ld1b { z0.b }, p3/Z, [x20, x28]\n"
    "umax z20.b, p4/M, z20.b, z24.b\n"
    "ld1b { z31.b }, p2/Z, [x23, x27]\n"
    "umax z19.b, p4/M, z19.b, z23.b\n"
    "ld1b { z30.b }, p2/Z, [x22, x27]\n"
    "umax z18.b, p4/M, z18.b, z22.b\n"
    "ld1b { z22.b }, p2/Z, [x21, x27]\n"
    "umax z17.b, p4/M, z17.b, z21.b\n"
    "ld1b { z29.b }, p2/Z, [x20, x27]\n"
    "umax z16.b, p4/M, z16.b, z20.b\n"
    "ld1b { z28.b }, p1/Z, [x23, x26]\n"
    "umax z10.b, p4/M, z10.b, z19.b\n"
    "ld1b { z27.b }, p1/Z, [x22, x26]\n"
    "umax z9.b, p4/M, z9.b, z18.b\n"
    "ld1b { z21.b }, p1/Z, [x21, x26]\n"
    "umax z8.b, p4/M, z8.b, z17.b\n"
    "ld1b { z26.b }, p1/Z, [x20, x26]\n"
    "umax z7.b, p4/M, z7.b, z16.b\n"
    "ld1b { z16.b }, p0/Z, [x23, x25]\n"
    "ld1b { z25.b }, p0/Z, [x22, x25]\n"
    "ld1b { z20.b }, p0/Z, [x21, x25]\n"
    "ld1b { z24.b }, p0/Z, [x20, x25]\n"
    "bgt 2b\n"
    "3:"  // 4-vectors of channels: 4 inputs tail
    "movprfx z19, z3\n umax z19.b, p4/M, z19.b, z2.b\n"
    "movprfx z23, z1\n umax z23.b, p4/M, z23.b, z0.b\n"
    "movprfx z18, z31\n umax z18.b, p4/M, z18.b, z30.b\n"
    "umax z22.b, p4/M, z22.b, z29.b\n"
    "movprfx z17, z28\n umax z17.b, p4/M, z17.b, z27.b\n"
    "umax z21.b, p4/M, z21.b, z26.b\n"
    "umax z16.b, p4/M, z16.b, z25.b\n"
    "umax z20.b, p4/M, z20.b, z24.b\n"
    "umax z19.b, p4/M, z19.b, z23.b\n"
    "umax z18.b, p4/M, z18.b, z22.b\n"
    "umax z17.b, p4/M, z17.b, z21.b\n"
    "umax z16.b, p4/M, z16.b, z20.b\n"
    "umax z10.b, p4/M, z10.b, z19.b\n"
    "umax z9.b, p4/M, z9.b, z18.b\n"
    "umax z8.b, p4/M, z8.b, z17.b\n"
    "umax z7.b, p4/M, z7.b, z16.b\n"
    "4:"  // 4-vectors of channels: After loop
    "ands x20, %x[n_valid_cells], #0x3\n"
    "beq 6f\n"
    "5:"  // 4-vectors of channels: Single input loop
    "ldr x23, [x19], #0x8\n"
    "subs x20, x20, #0x1\n"
    "ld1b { z3.b }, p3/Z, [x23, x28]\n"
    "umax z10.b, p4/M, z10.b, z3.b\n"
    "ld1b { z31.b }, p2/Z, [x23, x27]\n"
    "ld1b { z28.b }, p1/Z, [x23, x26]\n"
    "umax z9.b, p4/M, z9.b, z31.b\n"
    "ld1b { z16.b }, p0/Z, [x23, x25]\n"
    "umax z8.b, p4/M, z8.b, z28.b\n"
    "umax z7.b, p4/M, z7.b, z16.b\n"
    "bgt 5b\n"
    "6:"  // 4-vectors of channels: Single input loop: End
    "mov z6.s, #0x0\n"
    "add x19, %x[quant_params], %[offsetof_qp_input_offset]\n"
    "ld1rw { z5.s }, p4/Z, [x19]\n"
    "mov z4.s, #0xff\n"
    "add x19, %x[quant_params], %[offsetof_qp_per_layer_mul]\n"
    ".inst 0x4508a951  // ushllb z17.h, z10.b, #0x0\n"
    "ld1rw { z3.s }, p4/Z, [x19]\n"
    "add x19, %x[quant_params], %[offsetof_qp_per_layer_left_shift]\n"
    ".inst 0x4508ad50  // ushllt z16.h, z10.b, #0x0\n"
    "ld1rw { z2.s }, p4/Z, [x19]\n"
    "add x19, %x[quant_params], %[offsetof_qp_per_layer_right_shift]\n"
    ".inst 0x4508a937  // ushllb z23.h, z9.b, #0x0\n"
    "ld1rw { z1.s }, p4/Z, [x19]\n"
    "add x19, %x[quant_params], %[offsetof_qp_output_offset]\n"
    ".inst 0x4508ad36  // ushllt z22.h, z9.b, #0x0\n"
    "ld1rw { z0.s }, p4/Z, [x19]\n"
    ".inst 0x4508a912  // ushllb z18.h, z8.b, #0x0\n"
    ".inst 0x4508ad15  // ushllt z21.h, z8.b, #0x0\n"
    ".inst 0x4508a8f4  // ushllb z20.h, z7.b, #0x0\n"
    ".inst 0x4508acf3  // ushllt z19.h, z7.b, #0x0\n"
    "neg z5.s, p4/M, z5.s\n"
    ".inst 0x459140bf  // saddwb z31.s, z5.s, z17.h\n"
    ".inst 0x459144b1  // saddwt z17.s, z5.s, z17.h\n"
    ".inst 0x459040be  // saddwb z30.s, z5.s, z16.h\n"
    ".inst 0x459044b0  // saddwt z16.s, z5.s, z16.h\n"
    ".inst 0x459740bd  // saddwb z29.s, z5.s, z23.h\n"
    ".inst 0x459744bc  // saddwt z28.s, z5.s, z23.h\n"
    ".inst 0x459640bb  // saddwb z27.s, z5.s, z22.h\n"
    ".inst 0x459644ba  // saddwt z26.s, z5.s, z22.h\n"
    ".inst 0x459240b9  // saddwb z25.s, z5.s, z18.h\n"
    ".inst 0x459244b2  // saddwt z18.s, z5.s, z18.h\n"
    ".inst 0x459540b8  // saddwb z24.s, z5.s, z21.h\n"
    ".inst 0x459544b7  // saddwt z23.s, z5.s, z21.h\n"
    ".inst 0x459440b6  // saddwb z22.s, z5.s, z20.h\n"
    ".inst 0x459444b5  // saddwt z21.s, z5.s, z20.h\n"
    ".inst 0x459340b4  // saddwb z20.s, z5.s, z19.h\n"
    ".inst 0x459344b3  // saddwt z19.s, z5.s, z19.h\n"
    ".inst 0x4482905f  // srshl z31.s, p4/M, z31.s, z2.s\n"
    ".inst 0x44829051  // srshl z17.s, p4/M, z17.s, z2.s\n"
    ".inst 0x4482905e  // srshl z30.s, p4/M, z30.s, z2.s\n"
    ".inst 0x44829050  // srshl z16.s, p4/M, z16.s, z2.s\n"
    ".inst 0x04a377ff  // sqrdmulh z31.s, z31.s, z3.s\n"
    ".inst 0x04a37631  // sqrdmulh z17.s, z17.s, z3.s\n"
    ".inst 0x04a377de  // sqrdmulh z30.s, z30.s, z3.s\n"
    ".inst 0x04a37610  // sqrdmulh z16.s, z16.s, z3.s\n"
    ".inst 0x4482903f  // srshl z31.s, p4/M, z31.s, z1.s\n"
    ".inst 0x44829031  // srshl z17.s, p4/M, z17.s, z1.s\n"
    ".inst 0x4482903e  // srshl z30.s, p4/M, z30.s, z1.s\n"
    ".inst 0x44829030  // srshl z16.s, p4/M, z16.s, z1.s\n"
    "add z31.s, z31.s, z0.s\n"
    "add z17.s, z17.s, z0.s\n"
    "add z30.s, z30.s, z0.s\n"
    "add z16.s, z16.s, z0.s\n"
    ".inst 0x4482905d  // srshl z29.s, p4/M, z29.s, z2.s\n"
    ".inst 0x4482905c  // srshl z28.s, p4/M, z28.s, z2.s\n"
    ".inst 0x4482905b  // srshl z27.s, p4/M, z27.s, z2.s\n"
    ".inst 0x4482905a  // srshl z26.s, p4/M, z26.s, z2.s\n"
    ".inst 0x04a377bd  // sqrdmulh z29.s, z29.s, z3.s\n"
    ".inst 0x04a3779c  // sqrdmulh z28.s, z28.s, z3.s\n"
    ".inst 0x04a3777b  // sqrdmulh z27.s, z27.s, z3.s\n"
    ".inst 0x04a3775a  // sqrdmulh z26.s, z26.s, z3.s\n"
    ".inst 0x4482903d  // srshl z29.s, p4/M, z29.s, z1.s\n"
    ".inst 0x4482903c  // srshl z28.s, p4/M, z28.s, z1.s\n"
    ".inst 0x4482903b  // srshl z27.s, p4/M, z27.s, z1.s\n"
    ".inst 0x4482903a  // srshl z26.s, p4/M, z26.s, z1.s\n"
    "add z29.s, z29.s, z0.s\n"
    "add z28.s, z28.s, z0.s\n"
    "add z27.s, z27.s, z0.s\n"
    "add z26.s, z26.s, z0.s\n"
    ".inst 0x44829059  // srshl z25.s, p4/M, z25.s, z2.s\n"
    ".inst 0x44829052  // srshl z18.s, p4/M, z18.s, z2.s\n"
    "smax z31.s, p4/M, z31.s, z6.s\n"
    "smax z17.s, p4/M, z17.s, z6.s\n"
    ".inst 0x04a37739  // sqrdmulh z25.s, z25.s, z3.s\n"
    ".inst 0x04a37652  // sqrdmulh z18.s, z18.s, z3.s\n"
    "smin z31.s, p4/M, z31.s, z4.s\n"
    "smin z17.s, p4/M, z17.s, z4.s\n"
    ".inst 0x44829039  // srshl z25.s, p4/M, z25.s, z1.s\n"
    ".inst 0x44829032  // srshl z18.s, p4/M, z18.s, z1.s\n"
    "smax z30.s, p4/M, z30.s, z6.s\n"
    "trn1 z17.h, z31.h, z17.h\n"
    "add z25.s, z25.s, z0.s\n"
    "add z18.s, z18.s, z0.s\n"
    ".inst 0x44829058  // srshl z24.s, p4/M, z24.s, z2.s\n"
    ".inst 0x44829057  // srshl z23.s, p4/M, z23.s, z2.s\n"
    "smin z30.s, p4/M, z30.s, z4.s\n"
    "smax z16.s, p4/M, z16.s, z6.s\n"
    ".inst 0x04a37718  // sqrdmulh z24.s, z24.s, z3.s\n"
    ".inst 0x04a376f7  // sqrdmulh z23.s, z23.s, z3.s\n"
    "smax z29.s, p4/M, z29.s, z6.s\n"
    "smin z16.s, p4/M, z16.s, z4.s\n"
    ".inst 0x44829038  // srshl z24.s, p4/M, z24.s, z1.s\n"
    ".inst 0x44829037  // srshl z23.s, p4/M, z23.s, z1.s\n"
    "smin z29.s, p4/M, z29.s, z4.s\n"
    "trn1 z16.h, z30.h, z16.h\n"
    "add z24.s, z24.s, z0.s\n"
    "add z23.s, z23.s, z0.s\n"
    "trn1 z16.b, z17.b, z16.b\n"
    "st1b { z16.b }, p3, [%x[outptr], x28]\n"
    ".inst 0x44829056  // srshl z22.s, p4/M, z22.s, z2.s\n"
    "incb x28, ALL, MUL #4\n"
    ".inst 0x44829055  // srshl z21.s, p4/M, z21.s, z2.s\n"
    ".inst 0x44829054  // srshl z20.s, p4/M, z20.s, z2.s\n"
    ".inst 0x44829053  // srshl z19.s, p4/M, z19.s, z2.s\n"
    "smax z28.s, p4/M, z28.s, z6.s\n"
    ".inst 0x04a376d6  // sqrdmulh z22.s, z22.s, z3.s\n"
    ".inst 0x04a376b5  // sqrdmulh z21.s, z21.s, z3.s\n"
    ".inst 0x04a37694  // sqrdmulh z20.s, z20.s, z3.s\n"
    ".inst 0x04a37673  // sqrdmulh z19.s, z19.s, z3.s\n"
    ".inst 0x44829036  // srshl z22.s, p4/M, z22.s, z1.s\n"
    ".inst 0x44829035  // srshl z21.s, p4/M, z21.s, z1.s\n"
    ".inst 0x44829034  // srshl z20.s, p4/M, z20.s, z1.s\n"
    ".inst 0x44829033  // srshl z19.s, p4/M, z19.s, z1.s\n"
    "add z22.s, z22.s, z0.s\n"
    "add z21.s, z21.s, z0.s\n"
    "add z20.s, z20.s, z0.s\n"
    "add z19.s, z19.s, z0.s\n"
    "smax z27.s, p4/M, z27.s, z6.s\n"
    "smax z26.s, p4/M, z26.s, z6.s\n"
    "smax z25.s, p4/M, z25.s, z6.s\n"
    "smin z28.s, p4/M, z28.s, z4.s\n"
    "smin z27.s, p4/M, z27.s, z4.s\n"
    "smin z26.s, p4/M, z26.s, z4.s\n"
    "smin z25.s, p4/M, z25.s, z4.s\n"
    "trn1 z17.h, z29.h, z28.h\n"
    "smax z18.s, p4/M, z18.s, z6.s\n"
    "trn1 z16.h, z27.h, z26.h\n"
    "smax z24.s, p4/M, z24.s, z6.s\n"
    "trn1 z16.b, z17.b, z16.b\n"
    "st1b { z16.b }, p2, [%x[outptr], x27]\n"
    "smin z18.s, p4/M, z18.s, z4.s\n"
    "incb x27, ALL, MUL #4\n"
    "smin z24.s, p4/M, z24.s, z4.s\n"
    "smax z23.s, p4/M, z23.s, z6.s\n"
    "smax z22.s, p4/M, z22.s, z6.s\n"
    "smax z21.s, p4/M, z21.s, z6.s\n"
    "trn1 z18.h, z25.h, z18.h\n"
    "smin z23.s, p4/M, z23.s, z4.s\n"
    "smin z22.s, p4/M, z22.s, z4.s\n"
    "smin z21.s, p4/M, z21.s, z4.s\n"
    "smax z20.s, p4/M, z20.s, z6.s\n"
    "trn1 z16.h, z24.h, z23.h\n"
    "smax z19.s, p4/M, z19.s, z6.s\n"
    "trn1 z17.h, z22.h, z21.h\n"
    "trn1 z16.b, z18.b, z16.b\n"
    "st1b { z16.b }, p1, [%x[outptr], x26]\n"
    "smin z20.s, p4/M, z20.s, z4.s\n"
    "incb x26, ALL, MUL #4\n"
    "smin z19.s, p4/M, z19.s, z4.s\n"
    "trn1 z16.h, z20.h, z19.h\n"
    "trn1 z16.b, z17.b, z16.b\n"
    "st1b { z16.b }, p0, [%x[outptr], x25]\n"
    "incb x25, ALL, MUL #4\n"
    "whilelt p0.b, x25, %x[n_channels]\n"
    "b.any 1b\n"
    "7:"  // Single vector of channels
    "whilelt p3.b, x28, %x[n_channels]\n"
    "b.none 14f\n"
    "8:"  // Single vector of channels: Loop
    "mov z10.b, #0x0\n"
    "mov x19, %x[inptrs]\n"
    "lsr x24, %x[n_valid_cells], #0x2\n"
    "cbz x24, 11f\n"
    "ldp x23, x22, [x19, #0x0]\n"
    "ldp x21, x20, [x19, #0x10]\n"
    "add x19, x19, #0x20\n"
    "subs x24, x24, #0x1\n"
    "ld1b { z3.b }, p3/Z, [x23, x28]\n"
    "ld1b { z2.b }, p3/Z, [x22, x28]\n"
    "ld1b { z1.b }, p3/Z, [x21, x28]\n"
    "ld1b { z0.b }, p3/Z, [x20, x28]\n"
    "beq 10f\n"
    "9:"  // Single vector of channels: Loop: 4 inputs loop
    "movprfx z19, z3\n umax z19.b, p4/M, z19.b, z2.b\n"
    "ldp x23, x22, [x19, #0x0]\n"
    "subs x24, x24, #0x1\n"
    "movprfx z23, z1\n umax z23.b, p4/M, z23.b, z0.b\n"
    "ldp x21, x20, [x19, #0x10]\n"
    "add x19, x19, #0x20\n"
    "umax z19.b, p4/M, z19.b, z23.b\n"
    "ld1b { z3.b }, p3/Z, [x23, x28]\n"
    "ld1b { z2.b }, p3/Z, [x22, x28]\n"
    "umax z10.b, p4/M, z10.b, z19.b\n"
    "ld1b { z1.b }, p3/Z, [x21, x28]\n"
    "ld1b { z0.b }, p3/Z, [x20, x28]\n"
    "bgt 9b\n"
    "10:"  // Single vector of channels: Loop: 4 inputs tail
    "movprfx z19, z3\n umax z19.b, p4/M, z19.b, z2.b\n"
    "movprfx z23, z1\n umax z23.b, p4/M, z23.b, z0.b\n"
    "umax z19.b, p4/M, z19.b, z23.b\n"
    "umax z10.b, p4/M, z10.b, z19.b\n"
    "11:"  // Single vector of channels: Loop: After loop
    "ands x20, %x[n_valid_cells], #0x3\n"
    "beq 13f\n"
    "12:"  // Single vector of channels: Loop: Single input loop
    "ldr x23, [x19], #0x8\n"
    "subs x20, x20, #0x1\n"
    "ld1b { z3.b }, p3/Z, [x23, x28]\n"
    "umax z10.b, p4/M, z10.b, z3.b\n"
    "bgt 12b\n"
    "13:"  // Single vector of channels: Loop: Single input loop: End
    "mov z6.s, #0x0\n"
    "add x19, %x[quant_params], %[offsetof_qp_input_offset]\n"
    "ld1rw { z5.s }, p4/Z, [x19]\n"
    "mov z4.s, #0xff\n"
    "add x19, %x[quant_params], %[offsetof_qp_per_layer_mul]\n"
    ".inst 0x4508a951  // ushllb z17.h, z10.b, #0x0\n"
    "ld1rw { z3.s }, p4/Z, [x19]\n"
    "add x19, %x[quant_params], %[offsetof_qp_per_layer_left_shift]\n"
    ".inst 0x4508ad50  // ushllt z16.h, z10.b, #0x0\n"
    "ld1rw { z2.s }, p4/Z, [x19]\n"
    "add x19, %x[quant_params], %[offsetof_qp_per_layer_right_shift]\n"
    "neg z5.s, p4/M, z5.s\n"
    "ld1rw { z1.s }, p4/Z, [x19]\n"
    "add x19, %x[quant_params], %[offsetof_qp_output_offset]\n"
    ".inst 0x459140bf  // saddwb z31.s, z5.s, z17.h\n"
    "ld1rw { z0.s }, p4/Z, [x19]\n"
    ".inst 0x459144b1  // saddwt z17.s, z5.s, z17.h\n"
    ".inst 0x459040be  // saddwb z30.s, z5.s, z16.h\n"
    ".inst 0x459044b0  // saddwt z16.s, z5.s, z16.h\n"
    ".inst 0x4482905f  // srshl z31.s, p4/M, z31.s, z2.s\n"
    ".inst 0x44829051  // srshl z17.s, p4/M, z17.s, z2.s\n"
    ".inst 0x4482905e  // srshl z30.s, p4/M, z30.s, z2.s\n"
    ".inst 0x44829050  // srshl z16.s, p4/M, z16.s, z2.s\n"
    ".inst 0x04a377ff  // sqrdmulh z31.s, z31.s, z3.s\n"
    ".inst 0x04a37631  // sqrdmulh z17.s, z17.s, z3.s\n"
    ".inst 0x04a377de  // sqrdmulh z30.s, z30.s, z3.s\n"
    ".inst 0x04a37610  // sqrdmulh z16.s, z16.s, z3.s\n"
    ".inst 0x4482903f  // srshl z31.s, p4/M, z31.s, z1.s\n"
    ".inst 0x44829031  // srshl z17.s, p4/M, z17.s, z1.s\n"
    ".inst 0x4482903e  // srshl z30.s, p4/M, z30.s, z1.s\n"
    ".inst 0x44829030  // srshl z16.s, p4/M, z16.s, z1.s\n"
    "add z31.s, z31.s, z0.s\n"
    "add z17.s, z17.s, z0.s\n"
    "add z30.s, z30.s, z0.s\n"
    "add z16.s, z16.s, z0.s\n"
    "smax z31.s, p4/M, z31.s, z6.s\n"
    "smax z17.s, p4/M, z17.s, z6.s\n"
    "smax z30.s, p4/M, z30.s, z6.s\n"
    "smax z16.s, p4/M, z16.s, z6.s\n"
    "smin z31.s, p4/M, z31.s, z4.s\n"
    "smin z17.s, p4/M, z17.s, z4.s\n"
    "smin z30.s, p4/M, z30.s, z4.s\n"
    "smin z16.s, p4/M, z16.s, z4.s\n"
    "trn1 z17.h, z31.h, z17.h\n"
    "trn1 z16.h, z30.h, z16.h\n"
    "trn1 z16.b, z17.b, z16.b\n"
    "st1b { z16.b }, p3, [%x[outptr], x28]\n"
    "incb x28\n"
    "whilelt p3.b, x28, %x[n_channels]\n"
    "b.any 8b\n"
    "14:"  // End

    :
    : [inptrs] "r" (inptrs), [n_channels] "r" (n_channels), [n_valid_cells] "r" (n_valid_cells), [offsetof_qp_input_offset] "I" (offsetof(Requantize32, input_offset)), [offsetof_qp_output_offset] "I" (offsetof(Requantize32, output_offset)), [offsetof_qp_per_layer_left_shift] "I" (offsetof(Requantize32, per_layer_left_shift)), [offsetof_qp_per_layer_mul] "I" (offsetof(Requantize32, per_layer_mul)), [offsetof_qp_per_layer_right_shift] "I" (offsetof(Requantize32, per_layer_right_shift)), [outptr] "r" (outptr), [quant_params] "r" (&qp)
    : "cc", "memory", "p0", "p1", "p2", "p3", "p4", "x19", "x20", "x21", "x22", "x23", "x24", "x25", "x26", "x27", "x28", "z0", "z1", "z2", "z3", "z4", "z5", "z6", "z7", "z8", "z9", "z10", "z16", "z17", "z18", "z19", "z20", "z21", "z22", "z23", "z24", "z25", "z26", "z27", "z28", "z29", "z30", "z31"
  );
}

}  // namespace pooling
}  // namespace arm_conv

#endif  // defined(ARM_COMPUTE_ENABLE_SVE) && defined(ARM_COMPUTE_ENABLE_SVE2)
